/* ---------------------------------------------------------------- */
/* Copyright (c) Atelier de BioInformatique			    */
/* @file: fasta_io.c						    */
/* @desc: sequence IO in fasta format				    */
/*								    */
/* @history:							    */
/* @+	    <Gloup> : Aug 92 : first version			    */
/* @+       <Gloup> : Nov 95 : last revised version		    */
/* ---------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>       /*myerror*/

#define NEW(typ)                (typ*)malloc(sizeof(typ)) 
#define NEWN(typ, dim)          (typ*)malloc((unsigned)(dim) * sizeof(typ))
#define REALLOC(typ, ptr, dim)  (typ*)realloc((void *) (ptr), (unsigned long)(dim) * sizeof(typ))
#define NEWY(var, typ, dim)  {var=(typ*)malloc((size_t)(dim)*sizeof(typ));\
                                      if((var)==NULL){\
                                        fprintf(stderr,"pb malloc avec %s (%d %s) (size=%d)\n",\
                                        #var,__LINE__,__FILE__,(int)dim);exit(99);}} 
#define TREALLOC(typ, var, dim)  {var=(typ*)realloc((void *)(var),(size_t)(dim)*sizeof(typ));\
                                      if((var)==NULL){\
                                        fprintf(stderr,"pb realloc avec %s (%d %s)\n",\
                                        #var,__LINE__,__FILE__);exit(99);}} 
#define FREE(ptr)               free(ptr)                                                                                           


#ifdef SunOS
#include <unistd.h>
#endif

#include "Gtypes.h"
#include "libfasta.h"

#define SECURE   1	/* 1 : secure copy ; 0 : quickest copy 	*/

#define READ_NEXT Faux
#define PUSH_BACK Vrai

#define SERIAL    Vrai
#define INDEXED   Faux

#ifdef MACINTOSH
#define LINE_FEED '\r'
#else
#define LINE_FEED '\n'
#endif

#define INIT_SEQ_BUF	256	/* buffer initialement alloue pour une
				   sequence, 256 est bon pour les
				   proteines (double si manque
				   de place
				 */
static Bool sRetained[2] = {Faux,Faux};

/* -------------------------------------------- */
/* @static: lecture bufferisee			*/
/* -------------------------------------------- */
static char * sNextIOBuffer(FILE *streamin, Bool retain, Bool serial, int num)
{
	char   *buf, *end;
	Bool   reused;
	
	static char sBuffer[2][BUFSIZ];	/* BUFSIZ in <stdio.h>	*/
	
	reused = (retain || sRetained[num]) && serial;

	sRetained[num] = retain;

	buf = (   reused 
		? sBuffer[num] 
		: fgets(sBuffer[num], sizeof(sBuffer[num]), streamin));	 	

	if (! buf)		  /* end of file	    */
	    return NULL;

	end = buf + strlen(buf) - 1;

	if (*end == LINE_FEED)	  /* remove trailing \n	    */
	   *end = '\000';

	return (  *buf || reused
	        ? buf
		: sNextIOBuffer(streamin, retain, serial, num));
}

#if SECURE
/* -------------------------------------------- */
/* compte le nombre de caracteres alpha	dans	*/
/* un buffer					*/
/* -------------------------------------------- */
static long sCountAlpha(char *buf)
{
	long count;
	
	for (count = 0 ; *buf ; buf++)
	    if (isalpha(*buf))
		count++;
	
	return count;
}
#endif

#if SECURE
/* -------------------------------------------- */
/* copy only alpha chars from s2 to s1		*/
/* -------------------------------------------- */
static char *sStrcpyAlpha(char *s1, char *s2)
{
	for( ; *s2 ; s2++)
	    if (isalpha(*s2))
	    	*s1++ = *s2;

	*s1 = '\000';

	return s1;
}
#endif

/* -------------------------------------------- */
/* skip to next space in buffer			*/
/* -------------------------------------------- */
static char *sNextSpace(char *buffer)
{
	for (; *buffer ; buffer++)
	   if (isspace((int)(*buffer)))
	   	return buffer;
	
	return NULL;
}

/* -------------------------------------------- */
/* returns sequence name (FASTA)		*/
/* -------------------------------------------- */
char *GetFastaName(char *buffer)
{
	char c[2];
	static char name[FASTA_NAMLEN];

	*c = buffer[FASTA_NAMLEN];
	 buffer[FASTA_NAMLEN] = '\000';

	if (sscanf(buffer + 1, "%s", name) != 1)
	    (void) strcpy(name, "<no Name>");

	buffer[FASTA_NAMLEN] = *c;
	 
	return name;
}

/* -------------------------------------------- */
/* returns sequence comment (FASTA)		*/
/* -------------------------------------------- */
char *GetFastaComment(char *buffer)
{
	char   *space;
	static char comment[FASTA_COMLEN];

	buffer[FASTA_COMLEN] = '\000';
	
	space = sNextSpace(buffer);
	
	(void) strcpy(comment, (space ? space + 1 : "\000"));

	return comment;
}

/* -------------------------------------------- */
/* liberation d'une sequence			*/
/* -------------------------------------------- */
FastaSequencePtr FreeFastaSequence(FastaSequencePtr seq)
{
	if (seq) {
	    if (seq->seq)  FREE(seq->seq);
	    FREE(seq);
	}

	return NULL;
}
	
/* -------------------------------------------- */
/* allocation d'une sequence			*/
/* -------------------------------------------- */
FastaSequencePtr NewFastaSequence(void)
{
	FastaSequencePtr seq;
	
	if (! (seq = NEW(FastaSequence)))
	    return NULL;
	   
	seq->length   = 0;
	seq->offset   = -1;

	if (! (seq->seq = NEWN(char,  INIT_SEQ_BUF)))
	    return FreeFastaSequence(seq);

	seq->bufsize = INIT_SEQ_BUF;

	*(seq->name)    = '\000';
	*(seq->comment) = '\000';

	seq->ok = Vrai;
	
	return seq;
}

/* -------------------------------------------- */
/* lecture/redimensionnement d'une sequence au	*/
/* format Fasta					*/
/* <Lecture Serie>				*/
/* returns : Faux -> last sequence		*/
/*	     Vrai -> more to read		*/
/*           <but> you must check seq->ok !	*/
/*	     that may be 'Faux' (memory error)	*/
/* -------------------------------------------- */
Bool ReadFastaSequence(FILE *streamin, FastaSequencePtr seq, int num)
{
	long 	readlen, buflen;
	char 	*buffer, *tbuf;

	seq->ok = Faux;				/* assume error		*/

	buflen = seq->length = 0; 
	
	seq->offset = ftell(streamin);

	buffer = sNextIOBuffer(streamin, READ_NEXT, SERIAL, num);

	if (! (buffer && (*buffer == '>'))) 	/* sync error		*/
	    return Faux;			/* last sequence	*/
	
	if (seq->offset)
	    seq->offset -= (strlen(buffer) + 1);

	(void) strcpy(seq->name,    GetFastaName(buffer));
	
	(void) strcpy(seq->comment, GetFastaComment(buffer));
	
	while ((buffer = sNextIOBuffer(streamin, READ_NEXT, SERIAL, num))) {

	    if (*buffer == '>') {
	    	(void) sNextIOBuffer(streamin, PUSH_BACK, SERIAL, num); /* push it back */
		break;
	    }

#if SECURE	    
	    readlen = sCountAlpha(buffer);
#else
	    readlen = (long)strlen(buffer);
#endif
	
	    buflen +=  readlen;
   
	    if (buflen >= seq->bufsize) {
	    
	    	if (! (tbuf = REALLOC(char, seq->seq, 2 * buflen + 1)))
	    	   return Vrai;			/* but seq->ok is Faux	*/

		seq->seq = tbuf;
		
		seq->bufsize = 2 * buflen + 1;
		
	    }		
#if SECURE
	    sStrcpyAlpha(seq->seq + seq->length, buffer);
#else
	    (void) memcpy(seq->seq + seq->length, buffer, readlen);
#endif
	
	    seq->length = buflen;
	
	}

	seq->seq[seq->length] = '\000';

	return (seq->ok = Vrai);
}

/* -------------------------------------------- */
/* lecture/redimensionnement d'une sequence au	*/
/* format Fasta					*/
/* <Lecture Indexee>				*/
/* returns : Faux -> memory error		*/
/*	     Vrai -> read ok			*/
/* -------------------------------------------- */
Bool GetFastaSequence(FILE *streamin, FastaSequencePtr seq, int num)
{
	long 	readlen, buflen;
	char 	*buffer, *tbuf;

	seq->ok = Faux;				/* assume error		*/

	buflen = seq->length = 0; 
	
	(void) fseek(streamin, seq->offset, SEEK_SET);

	buffer = sNextIOBuffer(streamin, READ_NEXT, INDEXED, num);

	if (! (buffer && (*buffer == '>'))) 	/* sync error		*/
	    return Faux;			/* last sequence	*/
	
	if (seq->offset)
	    seq->offset -= (strlen(buffer) + 1);

	(void) strcpy(seq->name,    GetFastaName(buffer));
	
	(void) strcpy(seq->comment, GetFastaComment(buffer));
	
	while ((buffer = sNextIOBuffer(streamin, READ_NEXT, INDEXED, num))) {

	    if (*buffer == '>')
		break;

#if SECURE	    
	    readlen = sCountAlpha(buffer);
#else
	    readlen = (long)strlen(buffer);
#endif
	
	    buflen +=  readlen;
   
	    if (buflen >= seq->bufsize) {
	    
	    	if (! (tbuf = REALLOC(char, seq->seq, 2 * buflen + 1)))
	    	   return Faux;			

		seq->seq = tbuf;
		
		seq->bufsize = 2 * buflen + 1;
		
	    }		
#if SECURE
	    sStrcpyAlpha(seq->seq + seq->length, buffer);
#else
	    (void) memcpy(seq->seq + seq->length, buffer, readlen);
#endif
	
	    seq->length = buflen;
	
	}

	seq->seq[seq->length] = '\000';

	return (seq->ok = Vrai);
}

/* -------------------------------------------- */
/* ecriture d'une sequence au format Fasta	*/
/* -------------------------------------------- */
void WriteFastaSequence(FILE *streamou, FastaSequencePtr seq, 
			int char_per_line, int only_seq)
{
	long  i, nlines, rest;
	char *buf, *end, tempo;

	if (only_seq != ONLY_SEQ) {
	   (void) fputc('>', streamou);
	   (void) fputs((*(seq->name)    ? seq->name    : "no_name"), streamou);
	   (void) fputc(' ', streamou);
	   (void) fputs((*(seq->comment) ? seq->comment : "\000"), streamou);
	   (void) fputc(LINE_FEED, streamou);
	}

	nlines = seq->length / char_per_line;

	buf = seq->seq;

	for (i = 0 ; i < nlines ; i++) {
	    end = buf + char_per_line;
	    tempo = *end;
	    *end = '\000';
	    (void) fputs(buf, streamou);
	    (void) fputc(LINE_FEED , streamou);
	    *end = tempo;
	    buf += char_per_line;
	}

	if ((rest = (seq->length % char_per_line)) != 0) {
	   end = buf + rest;
	   tempo = *end;
	   *end = '\000';
	   (void) fputs(buf, streamou);
	   (void) fputc(LINE_FEED , streamou);
	   *end = tempo;
	}
}

/* -------------------------------------------- */
/* rewind db file				*/
/* -------------------------------------------- */
void RewindFastaDB(FILE *streamin, int num)
{
    sRetained[num] = Faux;	/* forget previous buffer  */
    
    if (streamin)
	rewind(streamin);
}


/*************** Erreurs **************************************/
void MyFastaError(FILE *stream, int stat, char *fmt, ...)
{
   va_list args;

   va_start(args,fmt);
   (void)vfprintf(stream, fmt, args);
   (void)fputc((int)'\n', stream);
   va_end(args);
   (void)fflush(stream);

   if (stat > 0)
      exit(stat);

   return;

}


void Fasta2Binaire(FILE *fich_binaire, FastaSequence *seq) {

    int nbitems;
    size_t taille_struct =  sizeof(FastaSequence);
    size_t taille_seq;

    nbitems = (int)(fwrite(seq,  (size_t)(taille_struct),(size_t)(1), fich_binaire));        
    if(nbitems < 1)
       MyFastaError(stderr, 1, "probleme ecriture binaire\n");     
    taille_seq = (int)(seq->length);
    nbitems = (int)(fwrite(seq->seq, (size_t)sizeof(char),(size_t)(taille_seq), fich_binaire));        
    if(nbitems < taille_seq)
       MyFastaError(stderr, 1, "probleme ecriture binaire\n");

    return;    
} 
    
    
    
    
    
FastaSequence *Binaire2Fasta (FILE *fich_fasta) {


    FastaSequence *seq;
    int nbitems;
    size_t taille_struct =  sizeof(FastaSequence);
    size_t taille_seq;

    if ((seq = NewFastaSequence()) == NULL) {
             fprintf(stderr,"not enought memory for seq (fasta)  ..\n");
             return NULL;
    }
  
    if ((nbitems = fread(seq, (size_t)(taille_struct),1, fich_fasta)) < 1) {
            if(feof(fich_fasta))
               fprintf(stderr,"fin du fichier !!! \n");
            else
               fprintf(stderr,"1 IO error 1 with sequence #%s\n", seq->name);
            FreeFastaSequence(seq);
            return NULL;   
    }
         
    taille_seq = (size_t)(seq->length);
    NEWY(seq->seq , char,(taille_seq+1)); 

    if ((nbitems = fread(seq->seq, (size_t)sizeof(char), (size_t)(taille_seq),fich_fasta)) < (size_t)(taille_seq)) {
        fprintf(stderr,"2 IO error 1 with sequence #%s %d \n", seq->name, nbitems);
        FreeFastaSequence(seq);
        return NULL;   
    }

    seq->seq[taille_seq] = '\0';
 
    return seq;
}
