/* -------------------------------------------- */
/* @file: libmtd_loglik.c	 			*/
/* $Id: libmtd_loglik.c,v 1.3 2005/02/11 15:36:55 annela Exp $ 					*/
/* -------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mtd.h"

/* -------------------------------------------- */
/* new MTDJacobi				*/
/* -------------------------------------------- */
MTDJacobi *NewMTDJacobi(int order, int nstates) {

  MTDJacobi	*m;
  int		i,j;

  if (! (m = (MTDJacobi *)malloc(sizeof(MTDJacobi)))) {
    fprintf(stderr,"NewMTDJacobi: not enough memory for MTDJacobi\n");
    return NULL;
  }

  m->order = order;
  m->nstates = nstates;

  if (! (m->dphi = (REAL *)malloc((size_t)order*sizeof(REAL)))) {
     free(m);
     return NULL;
  }

  if (! (m->dq = (REAL **)malloc((size_t)nstates*sizeof(REAL *)))) {
     free(m);
     return NULL;
  }

  for (i = 0; i < nstates; i++) {
     if (! (m->dq[i] = (REAL *)malloc((size_t)nstates*sizeof(REAL)))) {
	for (j = 0; j < i; j++)
	   free(m->dq[j]);
	free(m->dphi);
	free(m);
	return NULL;
     }
  }

  return m;

}

/* -------------------------------------------- */
/* free MTDJacobi                               */
/* -------------------------------------------- */
void FreeMTDJacobi(MTDJacobi *m) {

  int		i;

  if (m) {
     if (m->dphi)
        free(m->dphi);
     if (m->dq) {
        for (i = 0; i < m->nstates; i++)
	   if (m->dq[i])
	      free(m->dq[i]);
	free(m->dq);
     }
     free(m);
  }

  return;

}


/* -------------------------------------------- */
/* new MTDLJacobi				*/
/* -------------------------------------------- */
MTDLJacobi *NewMTDLJacobi(int order, int nstates) {

  MTDLJacobi	*m;

  if (! (m = (MTDLJacobi *)malloc(sizeof(MTDLJacobi)))) {
    fprintf(stderr,"NewMTDLJacobi: not enough memory for MTDLJacobi\n");
    return NULL;
  }

  m->order = order;
  m->nstates = nstates;
  m->sizetot = order + nstates*nstates + 1 + nstates;
  m->muphi = order + nstates*nstates;
  m->muq = m->muphi + 1;

  if (! (m->d = (REAL *)malloc((size_t)m->sizetot*sizeof(REAL)))) {
     free(m);
     return NULL;
  }

  return m;

}

/* -------------------------------------------- */
/* free MTDLJacobi                               */
/* -------------------------------------------- */
void FreeMTDLJacobi(MTDLJacobi *m) {

  if (m) {
     if (m->d)
        free(m->d);
     free(m);
  }

  return;

}


/* -------------------------------------------- */
/* compute log-likelihood of a single run of	*/
/* data						*/
/* if LAG_START <> 0 then  apply model on start	*/
/* of lag too					*/
/* -------------------------------------------- */

REAL logLSingle(INDX *index, int length, MTDModel *model) {

  REAL sum, ll;
  INDX *start, *pos, *end, *lag;

#if LAG_START
  INDX *lstart;
#endif

  start  = index;
  end    = start + length - 1;

#if LAG_START
  lstart = start + model->order;
  pos    = start + 1;
#else
  pos = start + model->order;
#endif

  ll = 0.;




  while (pos <= end) {

    sum = 0.;
    for (lag = start ; lag < pos ; lag++){
      sum += model->phi[pos-lag-1] * model->q[(int)(*lag)][(int)(*pos)];
    }
    if (sum <= 0.){
      sum = EPSILON;
    }
/*     fprintf(stderr, "LlogLSingle: proba %g, log: %g \n",  */
/* 	    model->phi[pos-lag-1] * model->q[(int)(*lag)][(int)(*pos)], */
/* 	    LOG(model->phi[pos-lag-1] * model->q[(int)(*lag)][(int)(*pos)])); */
    ll += LOG(sum);

#if LAG_START
    if (pos >= lstart) 
      start++;
#else
    start++;
#endif

    pos++;
  }


 return ll;
}

 
/* -------------------------------------------- */
/* compute log-likelihood of a single run of	*/
/* data						*/
/* if LAG_START <> 0 then  apply model on start	*/
/* of lag too					*/
/* -------------------------------------------- */

REAL ProbSingle(int *index, int length, MTDModel *model) {

  REAL sum, llProba=1, ll=0;
  int *start, *pos, *end, *lag;

  /* LAG_START is zero	*/
#if LAG_START
  INDX *lstart;
#endif

  start  = index;
  end    = start + length - 1;

#if LAG_START
  lstart = start + model->order;
  pos    = start + 1;
#else
  pos = start + model->order;
#endif





  while (pos <= end) {

    sum = 0.;
    /* computes proba(pos) from pairs from p(pos|start) to p(pos|pos-1) */
    for (lag = start ; lag < pos ; lag++){
      sum += model->phi[pos-lag-1] * model->q[(int)(*lag)][(int)(*pos)];
    }
    if (sum <= 0.){
      sum = EPSILON;
    }
    llProba=llProba*sum;
    ll += LOG(sum);
/*     fprintf(stderr, "LlogLSingle: proba angle %g, SHSP: %g log SHSP: %g \n",  */
/* 	    sum, llProba, ll); */

#if LAG_START
    if (pos >= lstart) 
      start++;
#else
    start++;
#endif

    pos++;
  }

/*MATH: Ne renvoie plus le log de la proba mais la proba elle mme pour pouvoir faire les "proba cumules"*/
/*    fprintf(stderr, "LlogLSingle: log(probaSHSPs): %g, proba SHSPs: %g\n", ll, llProba);*/

/*  return ll;*/
/*   fprintf(stderr, "ProbSingle:: proba du SHSP: %E, log proba: %g\n", llProba, ll); */
  return llProba;
}

/* -------------------------------------------- */
/* compute log-likelihood of data set		*/
/* -------------------------------------------- */

REAL logL(DataSet *data, MTDModel *model) {

  int i;
  REAL ll;

  ll  = 0.;

  for (i = 0 ; i < data->nseq ; i++) {
    ll += logLSingle(data->seq[i]->seq, data->seq[i]->length, model);
  }
  return ll;
}


/* -------------------------------------------- */
/* compute log-likelihood of a single run of	*/
/* data						*/
/* if LAG_START <> 0 then  apply model on start	*/
/* of lag too					*/
/* -------------------------------------------- */

REAL LlogLSingle(INDX *index, int length, MTDLModel *lmodel) {

  REAL sum, ll;
  INDX *start, *pos, *end, *lag;

#if LAG_START
  INDX *lstart;
#endif

  start  = index;
  end    = start + length - 1;

#if LAG_START
  lstart = start + lmodel->order;
  pos    = start + 1;
#else
  pos = start + lmodel->order;
#endif

  ll = 0.;

  while (pos <= end) {

    sum = 0.;
    for (lag = start ; lag < pos ; lag++)
      sum += lmodel->x[pos-lag-1] * lmodel->x[lmodel->order + (int)(*lag)*lmodel->nstates+(int)(*pos)];
      /* old jompo sum += model->phi[pos-lag-1] * model->q[(int)(*lag)][(int)(*pos)]; */
    if (sum <= 0.)
      sum = EPSILON;

    ll += LOG(sum);

#if LAG_START
    if (pos >= lstart) 
      start++;
#else
    start++;
#endif

    pos++;
  }

  return ll;
}

/* -------------------------------------------- */
/* compute log-likelihood of data set		*/
/* -------------------------------------------- */

REAL LlogL(DataSet *data, MTDLModel *model) {

  int i;
  REAL ll;

  ll  = 0.;

  for (i = 0 ; i < data->nseq ; i++) {
    ll += LlogLSingle(data->seq[i]->seq, data->seq[i]->length, model);
  }

  return ll;
}

/* -------------------------------------------- */
/* compute derivative of log-likelihood		*/
/* with respect to model parameters on a	*/
/* single sequence				*/
/*						*/
/* if LAG_START <> 0 then  apply model on start	*/
/* of lag too					*/
/* -------------------------------------------- */

static void sdlogL(SEQUENCEPTR seq, MTDModel *model,
		   MTDJacobi *jacob) {

  REAL sum;
  INDX *start, *pos, *end, *lag; 

#if LAG_START
  INDX *lstart;
#endif

  start = seq->seq;
  end   = start + seq->length - 1;

#if LAG_START
  lstart = start + model->order;
  pos    = start + 1;
#else
  pos = start + model->order;
#endif

  while (pos <= end) {

    sum = 0.;
    for (lag = start ; lag < pos ; lag++)
      sum += model->phi[pos-lag-1] * model->q[(int)(*lag)][(int)(*pos)];

    for (lag = start ; lag < pos ; lag++) {
      jacob->dphi[pos-lag-1] += (model->q[(int)(*lag)][(int)(*pos)] / sum);
      jacob->dq[(int)(*lag)][(int)(*pos)] += (model->phi[pos-lag-1] / sum);
    }

#if LAG_START
    if (pos >= lstart) 
      start++;
#else
    start++;
#endif

    pos++;
  }
}

/* -------------------------------------------- */
/* compute derivative of minus log-likelihood	*/
/* with respect to model parameters and		*/
/* lagrange multipliers on a single		*/
/* sequence					*/
/*						*/
/* if LAG_START <> 0 then  apply model on start	*/
/* of lag too					*/
/* -------------------------------------------- */

static void sdMinusLlogL(SEQUENCEPTR seq, MTDLModel *lmodel,
		   MTDLJacobi *ljacob) {

  REAL sum;
  INDX *start, *pos, *end, *lag; 

#if LAG_START
  INDX *lstart;
#endif

  start = seq->seq;
  end   = start + seq->length - 1;

#if LAG_START
  lstart = start + lmodel->order;
  pos    = start + 1;
#else
  pos = start + lmodel->order;
#endif

  while (pos <= end) {

    sum = 0.;
    for (lag = start ; lag < pos ; lag++)
      sum += lmodel->x[pos-lag-1] * lmodel->x[lmodel->order+(int)(*lag)*lmodel->nstates + (int)(*pos)];
      /* sum += model->phi[pos-lag-1] * model->q[(int)(*lag)][(int)(*pos)]; */

    for (lag = start ; lag < pos ; lag++) {
		/* dlog(L)/dphi[pos-lag-1]	*/
      ljacob->d[pos-lag-1] -= (lmodel->x[lmodel->order+(int)(*lag)*lmodel->nstates + (int)(*pos)] / sum);
		/* dlog(L)/dq[lag][pos]	*/
      ljacob->d[ljacob->order + (int)(*lag)*lmodel->nstates + (int)(*pos)] -=
	      (lmodel->x[pos-lag-1] / sum);
    /* OLD
      jacob->dphi[pos-lag-1] -= (model->q[(int)(*lag)][(int)(*pos)] / sum);
      jacob->dq[(int)(*lag)][(int)(*pos)] -= (model->phi[pos-lag-1] / sum);
     */
    }


#if LAG_START
    if (pos >= lstart) 
      start++;
#else
    start++;
#endif

    pos++;
  }
}

/* -------------------------------------------- */
/* compute derivative of log-likelihood		*/
/* with respect to model parameters on a	*/
/* dataset					*/
/* -------------------------------------------- */

void dlogL(DataSet *data, MTDModel *model, MTDJacobi *jacob) {

  int i, j;

  /* -------------------------- */
  /* reset jacobian		*/

  for (i = 0 ; i < model->order ; i++)
    jacob->dphi[i] = 0.;
   
  for (i = 0 ; i < model->nstates ; i++)
  for (j = 0 ; j < model->nstates ; j++)
    jacob->dq[i][j] = 0.;

  /* -------------------------- */
  /* compute new jacobian	*/

  for (i = 0 ; i < data->nseq ; i++)
    sdlogL(data->seq[i], model, jacob);

}


/* -------------------------------------------- */
/* compute derivative of MINUS log-likelihood	*/
/* with respect to model parameters and		*/
/* lagrange multiplier on a dataset		*/
/* -------------------------------------------- */

void dLlogL(DataSet *data, MTDLModel *lmodel, MTDLJacobi *ljacob) {

  int i, j;

  /* -------------------------- */
  /* reset jacobian		*/

  
  for (i = 0 ; i < lmodel->sizetot ; i++)
    ljacob->d[i] = 0.;
  

  /* -------------------------- */
  /* compute new jacobian	*/

  		/* compute df/dphi[i]=-dLogL(L)/dphi[i]	*/
  for (i = 0 ; i < data->nseq ; i++)
    sdMinusLlogL(data->seq[i], lmodel, ljacob);


  			/* lmodel->x[lmodel->muphi] is mu_phi	*/
  			/* lmodel->x[muq..muq+nstates-1] are mu_q[i]	*/

  		/* df/dphi[i]=-dLogL(L)/dphi[i]-mu_phi	*/
  for(i=0; i< ljacob->order; i++)
	    ljacob->d[i] -= lmodel->x[lmodel->muphi];

		/* df/q[i]=-dLogL(L)/dq[i] already computed by sdMinusLlogL	*/
  		/* now must sum with -mu_q[i]in order to have			*/
		/* df/q[i]=-dLogL(L)/dq[i]-mu_q[i]	*/
		/*	*/
		/* i in [ljacob->order, ljacob->muphi[ are the q[i][j]	*/
  for(i = ljacob->order; i < ljacob->muphi; i++)
         ljacob->d[i] -= lmodel->x[lmodel->muq + (i-ljacob->order)/ljacob->nstates];

  		/* df/d_mu_phi = -g(phi[]) = -(Sum(phi[i])-1)	*/
  		/*   = -Sum(phi[i]) + 1				*/
  ljacob->d[ljacob->muphi] = 1.0;
  for (i = 0; i < lmodel->order; i++)
    ljacob->d[ljacob->muphi] -= lmodel->x[i];

  		/* df/d_mu_q[i.] = -g(q[i,.]) = -(Sum_j(q[i,j])-1)	*/
  		/*   = -Sum_j(q[i,j]) + 1				*/
  for (i = 0; i < lmodel->nstates; i++) {
     ljacob->d[ljacob->muq+i] = 1.0;
     for (j = 0; j < lmodel->nstates; j++) 
	  ljacob->d[ljacob->muq+i] -= lmodel->x[lmodel->order + (lmodel->nstates*i)+j];

  }
  return;
}


