/* ------------------------------------------------------------------ */
/* Copyright (c) Atelier de BioInformatique                           */
/* @file: getSSR.c                                                    */
/* @desc: computes coverage of SSR on sequences and compares it 	  */
/*        with random sequences respecting the content of k-tuples of */
/*        the gene or of the set of genes     */
/* @history:                                                          */
/* @+	    <ed> : December 1998 : Created                            */
/*				: March 1999 seriously debugged                       */
/* ------------------------------------------------------------------ */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include "getSSR.h"
#include "st.h"
#include "libfasta.h"
#include "util.h"

#ifdef SUN
extern char     *optarg;
extern int      optind;
#endif

#ifdef SGI
#include <getopt.h>
#endif

/* -------------------------------------------- */
/* my power function			        */
/* -------------------------------------------- */

static int MyPow(int elt, int pow)
{
	int res, i;

	for (i=0, res=1; i<pow; i++)
		res=res*elt;

	return res;
}

/* -------------------------------------------- */
/* bug specific erreur			        */
/* -------------------------------------------- */
static void BugErreur(msg, stat)
	char	      *msg;
	int	      stat;
{
	char buffer[256];

	sprintf(buffer, "BUG: %s <Call Police>", msg);

	Error(buffer, stat);
}

/* -------------------------------------------- */
/* get float argument				*/
/* -------------------------------------------- */

static float sGetFArg(char *opt, char *arg)
{
    float   varg;
    char    buffer[BUFSIZ];

    if (   arg 
	&& (sscanf(arg, "%f", &varg) == 1)
	&& (varg >= 0.))
	return varg;

    (void) sprintf(buffer,  "[GetFArg] bad argument -%s %s",
			    opt, (arg ? arg : "<NULL>"));

    (void) Error(buffer, 10);

    return -1; /* never reached */
}

/* -------------------------------------------- */
/* get int argument				*/
/* -------------------------------------------- */

static int sGetIArg(char *opt, char *arg)
{
    int     varg;
    char    buffer[BUFSIZ];

    if (   arg 
	&& (sscanf(arg, "%d", &varg) == 1)
	&& (varg >= 0))
	return varg;

    (void) sprintf(buffer,  "[GetIArg] bad argument -%s %s",
			    opt, (arg ? arg : "<NULL>"));

    (void) Error(buffer, 10);

    return -1; /* never reached */
}

/* ------------------------------------------------ */
/* built frequency hashtable for k-1 length words   */
/* ------------------------------------------------ */
static int BuiltHashTable(hash, alphabet, ksize)
    HashTable	*hash;
    char	*alphabet;
    int		ksize;
{
    int	    i, bitsize;
    
    long   k, lk, pk;
    
    strcpy(hash->alphabet, alphabet);
    
    hash->ksize = ksize;
    
    hash->alphasize = strlen(alphabet);
    
    hash->tsum = 0;
    
		/* compute bitshift = log2(alphasize) */

    lk = 0;
    for (pk = 1 ; pk < hash->alphasize ; pk *= 2) 
		lk++;

    hash->bitshift = lk;
    hash->wsize    = pk;
   
    bitsize = (hash->ksize - 1) * hash->bitshift;
    
    if (bitsize >= sizeof(long) * 8)
		Error("k-uple too big with this alphabet", 20); 

		/* compute masks				*/
    lk = 0x1;
    for (i = 1 ; i < bitsize ; i++)
	lk = (lk << 1) | 0x1;
	
    hash->wmask = lk;

    lk = 0x1;
    for (i = 1 ; i < hash->bitshift ; i++)
	lk = (lk << 1) | 0x1;
	
    hash->smask = lk;
   
		/* compute table size  = wsize^(ksize-1) */
    
    pk = 1;
    for (i = 1 ; i < hash->ksize ; i++)
		pk *= hash->wsize;

    hash->tsize = pk;
    		
		/* allocate memory			 */
		
    hash->table = NEWN(int *, hash->tsize);
    
    if (! hash->table)
		Error("not enough memory for hash table", 10);
	
    for (k = 0 ; k < hash->tsize ; k++) {
		hash->table[k] = NEWN(int, hash->alphasize + 1);
		if (! hash->table[k])
	    	Error("not enough memory for hash table", 10);
		for (i = 0 ; i <= hash->alphasize ; i++)
	    	hash->table[k][i] = 0;
    }
    
    return 0;
}

/* ------------------------------------------------ */
/* free frequency hashtable			    */
/* ------------------------------------------------ */
static void FreeHashTable(hash)
    HashTable	*hash;
{
    int i;
    
    for (i = 1 ; i < hash->ksize ; i++)
	FREE(hash->table[i]);

    FREE(hash->table);
}

/* ------------------------------------------------ */
/* compute hash value				    */
/* ! entries are computed only on the k-1 first	    */
/* symbols					    */
/* ------------------------------------------------ */
static long HashValue(word, hash)
    char      *word;	/* encoded */
    HashTable *hash;
{
    int i, ii, h;
    
    ii = hash->ksize - 1;
    
    if (! ii)
	return 0;	/* special case for ksize == 1 */

    h = word[0];
    
    for (i = 1 ; i < ii ; i++) {
	h <<= hash->bitshift;  
	h |= (int) word[i];
    }
     
    return h;
}

/* ------------------------------------------------ */
/* Invert function: computes word (non encoded)	    */
/* returns NULL if not a word			    */
/* ------------------------------------------------ */
static char *ShaValue(h, hash)
    long     h;
    HashTable *hash;
{
    int	  i, j;
    char *b;
    
    static char sBuf[512];
    
    b = sBuf + hash->ksize - 2;
    
    for (i = 0 ; i < (hash->ksize - 1) ; i++, b--) {
		j = h & hash->smask;
		if (j >= hash->alphasize)
	    	return NULL;
		*b = hash->alphabet[j];
		h >>= hash->bitshift;
    }

    sBuf[hash->ksize - 1] = sBuf[hash->ksize] = '\000';

    return sBuf;
}

/* ------------------------------------------------ */
/* update hash value				    */
/* should be inline				    */
/* ------------------------------------------------ */
static long UpdateHashValue(old, newsymb, hash)
    long     old;
    int	      newsymb;	/* aka encoded char */
    HashTable *hash;
{
    return  (hash->ksize == 1) 
	   ? 0
	   : ((old << hash->bitshift) & hash->wmask) | newsymb;
}

/* ------------------------------------------------ */
/* reset frequency of k-uples to 0		    */
/* ------------------------------------------------ */
static void ResetFrequencies(HashTable *hash)
{
    int  i, j;

    for (i = 0 ; i < hash->tsize ; i++) {
		for (j = 0 ; j <= hash->alphasize ; j++)
	    	hash->table[i][j] = 0;
    }

    hash->tsum = 0;
}

/* ------------------------------------------------ */
/* encode sequence				    */
/* ------------------------------------------------ */
static void EncodeSequence(char *seq, HashTable *hash)
{
	int  pos;
	char *cseq;

	for (cseq = seq ; *cseq ; cseq++) {
	    pos = (int) (strchr(hash->alphabet, *cseq) - hash->alphabet);
	    *cseq = (char) pos;
	}

	*cseq = EOS; /* marker */
}

/* ------------------------------------------------ */
/* compute frequency of k-uples			    */
/* ------------------------------------------------ */
static void ComputeFrequencies(char *seq, HashTable *hash, int len)
{
    long  h, up;
    char   *cseq;
    
    if (len < hash->ksize)
	return;

    up = hash->alphasize;
    
    h = HashValue(seq, hash);
    
    for (cseq = seq + hash->ksize - 1; *cseq != EOS ; cseq++) {
   
		hash->table[h][*cseq]++;    /* this count   */
	
		hash->table[h][up]++;	    /* sum for h    */
	
		hash->tsum++;		    /* sum of all   */
	
		h = UpdateHashValue(h, *cseq, hash);
    }  
}

/* ------------------------------------------------ */
/* return uniform random number in [from, to[	    */
/* ------------------------------------------------ */
static long UniformRandom(from, to, ran)
    long   from, to;
    RanFunc ran;
{
    long i;
    
    i  = ran() * (to - from);	/* will truncate to integer part    */
				/* in [0, (to-from)[		    */
    i += from;			/* integer shift origin		    */

    if (i >= to)		/* very unlikely but == may appear  */
       i--;			/* from truncature error	    */
    
    return i;
}

/* ------------------------------------------------ */
/* generate random mono sequence		    */
/* exact case	    				    */
/* by permutations in shrinking 
/* ------------------------------------------------ */
static void ShuffleMonoRandom(char *seq, RanFunc ran,  int len)
{
    long i, j;
    char  *cseq, tmp;
    
    cseq = seq;
    
    for (i = 0 ; i < len ; i++) {

				/* -- draw integer in [i, seqlen[ -- */
	j = UniformRandom(i, len, ran);

				/* -- swap values ------------------ */
        tmp	= cseq[j];
	cseq[j] = cseq[i];
	cseq[i] = tmp;
    }
}

/* ------------------------------------------------ */
/* return non-uniform random number		    */
/* according to distribution in dist		    */
/* ------------------------------------------------ */
static Int32 DistribRandom(dist, siz, ran)
    Int32   *dist, siz;
    RanFunc ran;
{
    Int32 i, sum, dsum, unif;
    
    dsum = dist[siz];		/* where sum is stored	    */

				/* draw int. in [i, dsum]   */
    unif = UniformRandom(0, dsum+1, ran);

				/* search on repart.	    */
    for (i = sum = 0; i < siz ; i++) {
	sum += dist[i];
	if (unif <= sum)
	    return i;
    }
    
    BugErreur("(i >= siz) in DistribRandom", 22); 
    
    return (siz-1);
}
/* ------------------------------------------------ */
/* return non-uniform random number		    */
/* according to distribution in hash		    */
/* Note: a cleaner way is to :			    */
/* -1- compute (alphasize)^ksize distribution from  */
/*     hash->table values.			    */
/* -2- call DistribRandom			    */
/* but, this will requires an extra storage	    */
/* of (alphasize)^ksize Int32			    */
/* so here is a 'schmutziger' version :(	    */
/* ------------------------------------------------ */
static Int32 HashRandom(hash, ran)
    HashTable *hash;
    RanFunc ran;
{
    long i, j, sum, unif;
    
				/* draw int. in [i, tsum]   */
    unif = UniformRandom(0, hash->tsum+1, ran);

				/* search on repart.	    */
    for (i = sum = 0; i < hash->tsize ; i++) {
	for (j = 0 ; j < hash->alphasize ; j++) {
	    sum += hash->table[i][j];
	    if (unif <= sum)
		return (i*hash->alphasize + j);
	}
    }
    
    BugErreur("(i,j >= siz) in HashRandom", 22); 
    
    return (hash->tsize*hash->alphasize - 1);
}

/* ------------------------------------------------ */
/* generate random poly sequence                    */
/* approximate case (generate)                      */
/* ------------------------------------------------ */
static void GeneratePolyRandom(char *seq, HashTable *hash, 
	RanFunc ran, int len)
{
    Int32 i, j, h;
    char  *cseq, tmp;
    
    cseq = seq;
    
				/* -- choose initial ktuple -------- */
				
				/* choose a ktuple		     */
    i = HashRandom(hash, ran);	/* using kuples frequencies	     */
    
    j = i % hash->alphasize;	    /* kth symbol		     */

    h = (i-j) / hash->alphasize;    /* hash of first (k-1)uple	     */

				    /* copy first kuple		     */
    strncpy(cseq, ShaValue(h, hash), hash->ksize - 1);
    cseq[hash->ksize - 1] = '\000';		/* EOS for encoding  */
    EncodeSequence(seq, hash);			/*                   */
    cseq[hash->ksize - 1] = j;			/* kth symbol	     */
    
				    /* hash of second (k-1)tuple     */
    h = UpdateHashValue(h, j, hash);
      
				/* -- choose next ktuples ---------- */
				
    for (cseq = seq + hash->ksize ; *cseq != EOS ; cseq++) {
		*cseq = DistribRandom(hash->table[h], hash->alphasize, ran);
		h = UpdateHashValue(h, *cseq, hash);
    }  
}

/* ------------------------------------------------ */
/* generate random sequence	(allopt=0=> shuffle not exact)*/
/* ------------------------------------------------ */
static void GenerateRandom(char *seq, HashTable *hash, RanFunc ran, 
    int len, int allopt)
{
    if (len < hash->ksize)
		return;			    /* huh ?	    */

    if (hash->ksize == 1 && allopt==0)
		ShuffleMonoRandom(seq, ran, len);
    else
		GeneratePolyRandom(seq, hash, ran, len);
}

/* ------------------------------------------------ */
/* decode sequence				    */
/* ------------------------------------------------ */
static void DecodeSequence(char *seq, HashTable *hash)
{
	int  pos;
	char *cseq;

	for (cseq = seq ; *cseq != EOS ; cseq++)
	    *cseq = hash->alphabet[*cseq];

	*cseq = '\000';
}

/* ------------------------------------------------ */
/* compute comparative frequency tables		    */
/* and optionnally print it			    */
/* debug					    */
/* ------------------------------------------------ */

static double ComputeStatistics(datahash, shufhash, flen, allopt)
    HashTable *datahash, *shufhash;
	double flen;
	int allopt;
{
    long i, j, f2, s2, nu, last;
	double f1, s1, diff, sum, lchi, chi, prob;
    char  *buf;
    
    last = datahash->ksize-1;
    
    s1 = nu = 0;
	s2 = chi= 0.;
	
    for (i = 0 ; i < datahash->tsize ; i++) {

		if (! (buf = ShaValue(i, datahash)))
	    	continue;
			
		for (j = 0 ; j < datahash->alphasize ; j++) {
				
	    	buf[last] = datahash->alphabet[j];
	    
	    	f1 = datahash->table[i][j]*flen;
	    	f2 = shufhash->table[i][j];
	    	s1 += f1;
	    	s2 += f2;
	    	diff = f1 - f2;
	    	sum  = f1 + f2;

    	    nu++;
	    
	    	if (sum == 0) {
				lchi = 0.;
				nu--;
	    	}
	    	else
				lchi = diff * diff / sum;
	    
	    	chi += lchi;
	    }
    }
    
    prob = 1. - gammq((double)(0.5*(double)nu), (double)(0.5* chi));
        
    return prob;
}

/* ------------------------------------------------ */
/* print counts only   								*/
/* ------------------------------------------------ */
#define PP fprintf(stdout, 

static void outCounts(HashTable *hashres,
	MotifsTable *posit, int nposit, char *id, int nid,
	int nopt)
{
    long i, j, f1, s1, last;
    char  *buf;
	MotifsTable *pos;
    
    last = hashres->ksize-1;
        
    s1 = 0;

	for (i=0, pos = posit; i<nposit; i++, pos++){
		
		if (! pos->includ)	/* if it is a k-divisor*/
			continue;

		if (! (buf = ShaValue(pos->large, hashres)))
	    	continue;

		buf[last] = hashres->alphabet[pos->small];

	    f1 = hashres->table[pos->large][pos->small];
		s1 += f1;

		if ( nopt==0 || f1 != 0)  /* printing word, #, expected, chi2 */
			PP  "SR %s\t%ld\n", buf, f1); 	
 	}
 
 	if (nopt==0 || s1)  /*id, #serie, proba_chi2, obs, expected, observed, chi2 */
		PP  "SH %s %d %d\n//\n", id, nid, s1);
    
}

#undef PP

/* ------------------------------------------------ */
/* compute comparative frequency tables		    */
/* and  print it			    */
/* ------------------------------------------------ */
#define PP fprintf(stdout, 

static float ComputFinalStats(HashTable *hashres, HashTable *hash2,
	MotifsTable *posit, int nposit, int iter, char *id, int nid,
	int nopt)
{
    long i, j, f1, f2, s1, s2, nu, last;
	double diff, sum;
    float lchi, chi, prob;
    char  *buf, *aux;
	MotifsTable *pos;
    
    last = hashres->ksize-1;
        
    s1 = s2 = nu = chi = 0;

	for (i=0, pos = posit; i<nposit; i++, pos++){
		
		if (! pos->includ)	/* if it is a k-divisor*/
			continue;

		if (! (buf = ShaValue(pos->large, hashres)))
	    	continue;		

		buf[last] = hashres->alphabet[pos->small];

	    f1 = hashres->table[pos->large][pos->small];
	    f2 = hash2->table[pos->large][pos->small];
	    s1 += f1;
	    s2 += f2;
	    diff = f1 - f2/iter;
	    sum  = f1 + f2/iter;

    	nu++;

		if (sum < 0.00001){
 			lchi = 0.; 
			nu--;
			}
 	    else
 			lchi = (diff - 0.5) * (diff - 0.5) / sum; 
   
 	    chi += lchi;

		if ( nopt==0 || f1 != 0)  /* printing word, #, expected, chi2 */
			PP  "SR %s\t%ld\t%.2f\t%.2f\n", buf, f1, (float) f2/iter, lchi); 	
 	}

     prob = 1. - (float) gammq((double)(0.5*(float)nu), (double)(0.5* chi)); 
 
 	if (nopt==0 || s1)  /*id, #serie, proba_chi2, obs, expected, observed, chi2 */
		PP  "SH %s %d %.3f %ld %.2f %.2f\n//\n", id, nid, prob, s1,
			(float) s2/iter, chi);
    
    return prob;
}

#undef PP

/* -------------------------------------------- */
/* shuffle a sequence                           */
/* -------------------------------------------- */

static int shuffleSeq (double pval, HashTable *data_hash, 
	HashTable *shuf_hash, char *seq, int len, RanFunc rfunc,
	int allopt, double flen)
{
	
	int iter;
	double prob;
	
	prob = pval;

	iter = 0;
	
	if (data_hash->ksize == 1 && allopt==0)  /* since for 1 it is exact ... */
		GenerateRandom(seq, data_hash, rfunc, len, allopt);
	else {
    	do {
			iter++;    
		
			ResetFrequencies(shuf_hash);
			GenerateRandom(seq, data_hash, rfunc, len, allopt);
			ComputeFrequencies(seq, shuf_hash, len);
				
			prob = ComputeStatistics(data_hash, shuf_hash, flen, allopt);
/* 			fprintf(stderr, "iter: %d prob=%.2g pval=%f, factor=%g\n", 
				iter, prob, pval, flen);
 */	    
		} while ((prob > pval) && (iter <= MAX_PITER));
	}
	
	if (iter > MAX_PITER)
		return 0;
	
	return 1;
}

/* -------------------------------------------- */
/* get k-sizes below K that are divisors of k   */
/* -------------------------------------------- */

static int getdivid(int *divid, int ksiz)
{
	int count=0, i;
	
	if (ksiz==1)
		return 0;
	
	divid[count++]=1; /* 1 is always divisor */
	
	for (i=2; i<=ksiz/2; i++){
		if (ksiz%i == 0)
			divid[count++]=i;
	}
	
	return count;
}

/* -------------------------------------------- */
/* Fill structure with positions                */
/* -------------------------------------------- */

static void BuildPosit(MotifsTable *posit, int nposit, int ksiz, 
	int ssrsiz)
{
	int i, j, lsmal, llarj, step;
	MotifsTable *pos;
	
	pos = posit;
					/* first element */
	pos->large = 0;
	pos->small = 0;
	pos->includ = 1;	/* by default it includes all */
	
	for (i=1, step=0; i<= (ssrsiz/ksiz)-1; i++)
		step += MyPow(ALPHA_SIZ, (i*ksiz-1));
	
	for (i=1, lsmal=0, llarj=0; i<nposit; i++){
		pos++;
		llarj += step;

		if ( (lsmal+1)%ALPHA_SIZ )		/* circles around 4 */
			lsmal = lsmal+1;
		else {
			lsmal = 0;
			llarj ++;	/* cicled over */
		}
				
		pos->large = llarj;
		pos->small = lsmal;
		pos->includ= 1;
	}
}

/* -------------------------------------------- */
/* withdraw from posit the smaller k-tuples    */
/* -------------------------------------------- */

static void WithdrawPosit(MotifsTable *posit, int nposit, int ksiz)
{
	int i, j, step, divid[5], ndiv, nelt;
	MotifsTable *pos;
	
	ndiv = getdivid(divid, ksiz);	/* # of diff divisors*/
	
	for (i=0; i< ndiv; i++){
		nelt = MyPow(ALPHA_SIZ, divid[i]);	/* # to withdraw from this divisor */
		step = nposit/(nelt-1);
		for (j=0, pos = posit; j< nelt; j++, pos += step)
			pos->includ = 0;
	}
}

/* -------------------------------------------- */
/* printout header				*/
/* -------------------------------------------- */

#define PP fprintf(stdout, 

static void sPrintHeader(char *name, 
			 int argn, char *argv[], int ksiz, int ssrsiz, 
			 float proba, int nelts, int nopt, int niter, 
			 int allopt, int optK, int onlycnt)
{
    int	      i, j;

    PP	"XX ------------------------------------------			\n");
    PP	"XX  getSSR Version 1, December 98	\n");
    PP	"XX ------------------------------------------			\n");
    PP	"XX Command line     : ");
	
	for (i = 0 ; i < argn ; i++)
	    PP "%s ", argv[i]);
    
    PP	"\nXX\n");

    PP	"XX  k-tuple size    : %d\n", ksiz);
    PP	"XX  SSR size        : %d\n", ssrsiz);
    PP	"XX  # of words      : %d\n", nelts);

	if (onlycnt)
    	PP	"XX  Only perfming counts\n");
	else {
    	PP	"XX  shuffle pvalue  : %.5f\n", proba);
    	PP	"XX  # iterations    : %d\n", niter);
		if (allopt)
    		PP	"XX  counts taken from all genes \n");
		else
    		PP	"XX  counts taken from each gene \n");		
		}
	if (nopt)
    	PP	"XX  Not printing words with zero occurrences \n");
	if (optK)
    	PP	"XX  Not printing k-tuples of smaller complexity\n");
		
    PP	"XX\n");
}

#undef PP

/* -------------------------------------------- */
/* main stuff                                   */
/* -------------------------------------------- */

main( argn, argv)
	int  argn;
	char *argv[];
{

    int        i, nseq, ksize, seed, ssrsize, nposit, carg, errflag;
	int 	   niter, nopt, allopt, totlen, optK, onlycnt;
    double     pval, prob;		
    HashTable  data_hash, shuf_hash, shufres_hash, res_hash ;
    SeedFunc   sfunc;
    char       alphabet[MAX_SYMB], filin[256];
    RanFunc	   rfunc;
	FastaSequence *seq;
 	MotifsTable  *posit;
	
    pval = 0.001;	    /* default is 1 per 1000 */
    seed = (int) time(0) & 0xfffffff;
    sfunc = sranLE;
    rfunc = ranLE;
    sfunc(seed);
	niter = 100;
	allopt = 0;	/* default: use of gene counts */
	totlen = 0; /* length of sequences cumulated */
	
	strcpy(alphabet, ALPHA_DNA);	/* default DNA*/
    nseq=0;							
    ksize=1;						/* default in mono */
	
	ssrsize = 4;						/* default SSR size is 4 */
	errflag = 0;
	nopt = 0;							/* print zeros by default */
	optK = 0;						/* all k-tuples by default */
	onlycnt = 0;		/* by default count and shuffle */
	    
				/* -------------------- */
				/* parse arguments	*/
				/* -------------------- */

 	while ((carg = getopt(argn, argv, "achi:k:Knp:s:")) != -1) {

 	  switch(carg) {

        case 'a' :              /* stats come from all genes */
		    allopt = 1;
		    break;

        case 'c' :              /* counts only - do not shuffle */
		    onlycnt = 1;
		    break;

        case 'h' :              /* help             */
		    PrintHelp();
		    exit(0);

        case 'i' :              /* iter             */
		    niter = sGetIArg("i", optarg);
		    break;

		case 'k' :		/* ksize */
		    ksize = sGetIArg("k", optarg);
			if (ksize>10){
				Error("k-tuple size too large", 0);
				errflag++;
				}
		    break;

		case 'K' :		/* only of ksize (e.g. not AA for k=2) */
			optK = 1;
		    break;

		case 'n' :		/* do not print zeros */
		    nopt = 1;
		    break;

		case 'p' :		/* proba for iteration */
		    pval = sGetFArg("p", optarg);
		    if (pval > 1. || pval < 0.) {
				Error("p_value should be <= 1.", 0);
				errflag++;
		    }
		    break;

		case 's' :		/* SSR size */
		    ssrsize = sGetIArg("s", optarg);
			if (ssrsize>12){
				Error("ssr too large", 0);
				errflag++;
				}
		    break;

		case '?' :		/* misusage	    */
		    errflag++;
		    break;
	    }
	}
				/* -------------------- */
				/* check arguments	*/
				/* -------------------- */
				
	if ((argn - optind) > 1 || argn==1)  /* 1 argument */
	    errflag++;

	if (onlycnt)
		allopt = 0;
	
	if (errflag) 
	    ExitUsage(1);

			/* if 1 arg left -> sequence file */

	strcpy(filin, argv[optind]);
	
	AssignToStdin(filin);

	if (ssrsize%ksize!=0)
		Error("SSR size not a multiple of ksize",  1);
	
	if (! (seq = NewFastaSequence()))
		Error("Not enough memory",  10);

	nposit = MyPow(ALPHA_SIZ, ksize);
	
				/* creating the tables*/
	BuiltHashTable(&res_hash, alphabet, ssrsize);

	if (! onlycnt){		/* they're not required if only counting */
		BuiltHashTable(&data_hash, alphabet, ksize);
		BuiltHashTable(&shuf_hash, alphabet, ksize);
		BuiltHashTable(&shufres_hash, alphabet, ssrsize);	
	}

	if (! (posit = NEWN(MotifsTable, nposit+1)))
		Error("Not enough memory",  11);

	BuildPosit(posit, nposit, ksize, ssrsize);
	
	if (optK)		/* withdraw smaller k-tuples */
		WithdrawPosit(posit, nposit, ksize);
	
	sPrintHeader(filin, argn, argv, ksize, ssrsize, pval, nposit, 
		nopt, niter, allopt, optK, onlycnt);	
	
	if (allopt) {	/* all count taken together for shuffle stats*/

 		ResetFrequencies(&data_hash);
		
		while (ReadFastaSequence(stdin, seq)) {	

			if (! checkSeq(seq->seq, 1, alphabet))
				Error("Error in sequence", 1);

			totlen += seq->length;
			
					/* doing k-tuples and cumulating in data_hash */
			EncodeSequence(seq->seq, &data_hash);
			ComputeFrequencies(seq->seq, &data_hash, seq->length);
			DecodeSequence(seq->seq, &data_hash);
		}
		ReAssignToStdin(filin); /* reopening the file */
	}
	
    while (ReadFastaSequence(stdin, seq)) {		/* for each sequence*/

	    nseq++;

		if (! checkSeq(seq->seq, 1, alphabet))
			Error("Error in sequence", 1);
					
					/* doing observed SSR and putting in res_hash */
		ResetFrequencies(&res_hash);
		EncodeSequence(seq->seq, &res_hash);
		ComputeFrequencies(seq->seq, &res_hash, seq->length);
		
		if (onlycnt){			/* if just count stops here */
			outCounts(&res_hash, posit, nposit, seq->name, nseq, nopt);
			DecodeSequence(seq->seq, &res_hash);
			continue;			
		}
		
		if (! allopt) {    /* doing k-tuples and putting in data_hash */
			ResetFrequencies(&data_hash);
			ComputeFrequencies(seq->seq, &data_hash, seq->length);
		}
					
					/* shuffling to get expected */
		ResetFrequencies(&shufres_hash);
		
		for (i=0; i< niter; i++){ 

			if (! shuffleSeq(pval, &data_hash, &shuf_hash, seq->seq, 
					seq->length, rfunc, allopt, (double) seq->length/totlen)){
				fprintf(stderr, "Error in Shuffling: continuing... (%d)\n", i);
				continue;
			}

			ComputeFrequencies(seq->seq, &shufres_hash, seq->length);
		}
		
	/* remark: it presuposes that all alphabets are similar!!*/
	/* else it would require all the coding and decoding in between */
	
		DecodeSequence(seq->seq, &res_hash);
		ComputFinalStats(&res_hash, &shufres_hash, posit, nposit, 
			niter, seq->name, nseq, nopt);
    }
    
	FreeFastaSequence(seq);
	if (! onlycnt){		/* they're not required if only counting */
		FreeHashTable(&data_hash);
		FreeHashTable(&shufres_hash);
		FreeHashTable(&shuf_hash);
	}
	FreeHashTable(&res_hash);
	FREE(posit);

    exit(0);    
    
}
